import warnings

import numpy as np
from statsmodels.tsa.arima import model

_ARIMA_MODEL_DEPRECATION_WARNING = (
    "statsmodels.tsa.arima_model.Arima is deprecated "
    "and will be removed in a future version. Please use "
    "statsmodels.tsa.arima.model.ARIMA instead."
)


class ARIMA(object):
    """A Wrapper for the statsmodels.tsa.arima.model.ARIMA class."""

    def __init__(self, p, d, q, trend, steps):
        """Initialize the ARIMA object.

        Args:
            p (int):
                Integer denoting the order of the autoregressive model.
            d (int):
                Integer denoting the degree of differencing.
            q (int):
                Integer denoting the order of the moving-average model.
            trend (str):
                Parameter controlling the deterministic trend. Can be specified
                as a string where 'c' indicates a constant term, 't' indicates
                a linear trend in time, and 'ct' includes both.
            steps (int):
                Integer denoting the number of time steps to predict ahead.
        """
        warnings.warn(_ARIMA_MODEL_DEPRECATION_WARNING, DeprecationWarning)

        self.p = p
        self.d = d
        self.q = q
        self.trend = trend
        self.steps = steps

    def predict(self, X):
        """Predict values using the initialized object.

        Args:
            X (ndarray):
                N-dimensional array containing the input sequences for the model.

        Returns:
            ndarray:
                N-dimensional array containing the predictions for each input sequence.
        """
        arima_results = list()
        dimensions = len(X.shape)

        if dimensions > 2:
            raise ValueError("Only 1D o 2D arrays are supported")

        if dimensions == 1 or X.shape[1] == 1:
            X = np.expand_dims(X, axis=0)

        num_sequences = len(X)
        for sequence in range(num_sequences):
            arima = model.ARIMA(X[sequence], order=(self.p, self.d, self.q), trend=self.trend)
            arima_fit = arima.fit()
            arima_results.append(arima_fit.forecast(self.steps)[0])

        arima_results = np.asarray(arima_results)

        if dimensions == 1:
            arima_results = arima_results[0]

        return arima_results



import warnings

from mlstars.utils import import_object

_RESAMPLE_AGGS = [
    'mean',
    'median',
    'prod',
    'quantile',
    'std',
    'sum',
    'var',
]


def resample(df, rule, on=None, groupby=(), aggregation='mean',
             reset_index=True, time_index=None):
    """pd.DataFrame.resample adapter.

    Call the `df.resample` method on the given time_index
    and afterwards call the indicated aggregation.

    Optionally group the dataframe by the indicated columns before
    performing the resampling.

    If groupby option is used, the result is a multi-index datagrame.

    Args:
        df (pandas.DataFrame):
            DataFrame to resample.
        rule (str or int):
            The offset string or object representing target conversion or an
            integer value that will be interpreted as the number of seconds.
        on (str or None):
            Name of the column to use as the time index. If ``None`` is given, the
            DataFrame index is used.
        groupby (list):
            Optional list of columns to group by.
        aggregation (callable or str):
            Function or name of the function to use for the aggregation. If a name is given, it
            can either be one of the standard pandas aggregation functions or the fully qualified
            name of a python function that will be imported and used.
        reset_index (bool):
            Whether to reset the index after aggregating
        time_index (str or None):
            Deprecated: This has been renamed to `on`.
            Name of the column to use as the time index. If ``None`` is given, the
            DataFrame is index is used.

    Returns:
        pandas.Dataframe:
            resampled dataframe
    """
    if on is None and time_index is not None:
        message = (
            'resample `time_series` argument deprecated and will be removed'
            ' in future versions of ml-stars. Please use `on` instead.'
        )
        warnings.warn(message, DeprecationWarning, stacklevel=2)
        on = time_index

    if groupby:
        df = df.groupby(groupby)

    if isinstance(rule, int):
        rule = '{}s'.format(rule)

    dtir = df.resample(rule, on=on)

    if not callable(aggregation) and aggregation not in _RESAMPLE_AGGS:
        try:
            aggregation = import_object(aggregation)
        except (AttributeError, ImportError, ValueError):
            pass

    df = dtir.aggregate(aggregation)
    for name in df.index.names:
        if name in df:
            del df[name]

    if reset_index:
        df.reset_index(inplace=True)

    return df


def _join_names(names):
    """Join the names of a multi-level index with an underscore."""

    levels = (str(name) for name in names if name != '')
    return '_'.join(levels)


def unstack(df, level=-1, reset_index=True):
    """pd.DataFrame.unstack adapter.

    Call the `df.unstack` method using the indicated level and afterwards
    join the column names using an underscore.

    Args:
        df (pandas.DataFrame): DataFrame to unstack.
        level (str, int or list): Level(s) of index to unstack, can pass level name
        reset_index (bool): Whether to reset the index after unstacking

    Returns:
        pandas.Dataframe: unstacked dataframe
    """
    df = df.unstack(level=level)
    if reset_index:
        df = df.reset_index()
        df.columns = df.columns.map(_join_names)

    return df



from datetime import datetime
from unittest import TestCase

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from mlstars.adapters.pandas import resample


class ResampleTest(TestCase):

    def setUp(self):
        self.df = pd.DataFrame({
            'dt': [
                datetime(2000, 1, day + 1, hour, 0)
                for day in range(4)
                for hour in range(24)
            ],
            'value': np.arange(0, 4 * 24, dtype=float)
        })

    def test_resample_rule_str(self):

        out = resample(self.df.set_index('dt'), '1d')

        assert_frame_equal(out, pd.DataFrame([
            {'dt': datetime(2000, 1, 1), 'value': 11.5},
            {'dt': datetime(2000, 1, 2), 'value': 35.5},
            {'dt': datetime(2000, 1, 3), 'value': 59.5},
            {'dt': datetime(2000, 1, 4), 'value': 83.5},
        ]))

    def test_resample_rule_int(self):

        out = resample(self.df.set_index('dt'), 86400)

        assert_frame_equal(out, pd.DataFrame([
            {'dt': datetime(2000, 1, 1), 'value': 11.5},
            {'dt': datetime(2000, 1, 2), 'value': 35.5},
            {'dt': datetime(2000, 1, 3), 'value': 59.5},
            {'dt': datetime(2000, 1, 4), 'value': 83.5},
        ]))

    def test_resample_groupby(self):

        self.df['group1'] = ['A', 'B'] * 2 * 24
        self.df['group2'] = ['C', 'C', 'D', 'D'] * 24

        out = resample(self.df.set_index('dt'), '1d', groupby=['group1', 'group2'])

        assert_frame_equal(out, pd.DataFrame([
            {'group1': 'A', 'group2': 'C', 'dt': datetime(2000, 1, 1), 'value': 10.0},
            {'group1': 'A', 'group2': 'C', 'dt': datetime(2000, 1, 2), 'value': 34.0},
            {'group1': 'A', 'group2': 'C', 'dt': datetime(2000, 1, 3), 'value': 58.0},
            {'group1': 'A', 'group2': 'C', 'dt': datetime(2000, 1, 4), 'value': 82.0},
            {'group1': 'A', 'group2': 'D', 'dt': datetime(2000, 1, 1), 'value': 12.0},
            {'group1': 'A', 'group2': 'D', 'dt': datetime(2000, 1, 2), 'value': 36.0},
            {'group1': 'A', 'group2': 'D', 'dt': datetime(2000, 1, 3), 'value': 60.0},
            {'group1': 'A', 'group2': 'D', 'dt': datetime(2000, 1, 4), 'value': 84.0},
            {'group1': 'B', 'group2': 'C', 'dt': datetime(2000, 1, 1), 'value': 11.0},
            {'group1': 'B', 'group2': 'C', 'dt': datetime(2000, 1, 2), 'value': 35.0},
            {'group1': 'B', 'group2': 'C', 'dt': datetime(2000, 1, 3), 'value': 59.0},
            {'group1': 'B', 'group2': 'C', 'dt': datetime(2000, 1, 4), 'value': 83.0},
            {'group1': 'B', 'group2': 'D', 'dt': datetime(2000, 1, 1), 'value': 13.0},
            {'group1': 'B', 'group2': 'D', 'dt': datetime(2000, 1, 2), 'value': 37.0},
            {'group1': 'B', 'group2': 'D', 'dt': datetime(2000, 1, 3), 'value': 61.0},
            {'group1': 'B', 'group2': 'D', 'dt': datetime(2000, 1, 4), 'value': 85.0},
        ], columns=['group1', 'group2', 'dt', 'value']))

    def test_resample_on(self):

        out = resample(self.df, '1d', on='dt')

        assert_frame_equal(out, pd.DataFrame([
            {'dt': datetime(2000, 1, 1), 'value': 11.5},
            {'dt': datetime(2000, 1, 2), 'value': 35.5},
            {'dt': datetime(2000, 1, 3), 'value': 59.5},
            {'dt': datetime(2000, 1, 4), 'value': 83.5},
        ]))

    def test_resample_reset_index_false(self):

        out = resample(self.df.set_index('dt'), '1d', reset_index=False)

        assert_frame_equal(out.reset_index(), pd.DataFrame([
            {'dt': datetime(2000, 1, 1), 'value': 11.5},
            {'dt': datetime(2000, 1, 2), 'value': 35.5},
            {'dt': datetime(2000, 1, 3), 'value': 59.5},
            {'dt': datetime(2000, 1, 4), 'value': 83.5},
        ]))

    def test_resample_aggregation_str(self):

        out = resample(self.df.set_index('dt'), '1d', aggregation='max')

        assert_frame_equal(out, pd.DataFrame([
            {'dt': datetime(2000, 1, 1), 'value': 23.0},
            {'dt': datetime(2000, 1, 2), 'value': 47.0},
            {'dt': datetime(2000, 1, 3), 'value': 71.0},
            {'dt': datetime(2000, 1, 4), 'value': 95.0},
        ]))

    def test_resample_aggregation_func(self):

        out = resample(self.df.set_index('dt'), '1d', aggregation=np.max)

        assert_frame_equal(out, pd.DataFrame([
            {'dt': datetime(2000, 1, 1), 'value': 23.0},
            {'dt': datetime(2000, 1, 2), 'value': 47.0},
            {'dt': datetime(2000, 1, 3), 'value': 71.0},
            {'dt': datetime(2000, 1, 4), 'value': 95.0},
        ]))

    def test_resample_aggregation_import(self):

        out = resample(self.df.set_index('dt'), '1d', aggregation='numpy.max')

        assert_frame_equal(out, pd.DataFrame([
            {'dt': datetime(2000, 1, 1), 'value': 23.0},
            {'dt': datetime(2000, 1, 2), 'value': 47.0},
            {'dt': datetime(2000, 1, 3), 'value': 71.0},
            {'dt': datetime(2000, 1, 4), 'value': 95.0},
        ]))



# -*- coding: utf-8 -*-

from sklearn.preprocessing import MinMaxScaler


class Scaler(MinMaxScaler):
    """sklearn.preprocessing.MinMaxScaler adapter.

    Convert feature_range into a tuple.

    Args:
        feature_range (list):
            List of two elements. Desired range of transformed data.

        copy (bool):
            Set to False to perform inplace row normalization and avoid a copy.
    """

    def __init__(self, feature_range=(0, 1), copy=True):
        feature_range = tuple(feature_range)

        super(Scaler, self).__init__(
            feature_range=feature_range,
            copy=copy
        )
